/*
 *  savecore.c
 *
 *  Copyright (C) 2004, 2005  FUJITSU LIMITED
 *  Copyright (C) 2001-2004  Red Hat, Inc
 *
 *  Written by Nobuhiro Tachino (tachino@jp.fujitsu.com) based on
 *  netdump-server code.
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/* $Id: savecore.c,v 1.53 2005/08/08 19:14:01 tachino Exp $ */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/statfs.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#include "savecore.h"

#define CMD	"savecore"


#ifdef FORCE_SINGLE_PTLOAD
#define _FORCE_SINGLE_PTLOAD	true
#else
#define _FORCE_SINGLE_PTLOAD	false
#endif

static bool force_single_ptload	= _FORCE_SINGLE_PTLOAD;

static bool prog_flg = false;	/* becomes true if option '-p' specified. */
static bool frc_flg = false;	/* becomes true if option '-f' specified. */
static bool del_flg = false;	/* becomes true if option '-D' specified. */

static const int symbol = '#';	/* for progress bar */

static int nr_cpus;

gboolean generate_elf_header(DumpDevice *);


static DumpDevice *create_new_device(const char *device_name)
{
	DumpDevice *device;

	device = g_new0(DumpDevice, 1);

	device->device_name = g_strdup(device_name);

	device->file_fd = -1;
	device->device_fd = -1;

	/* Lame default. should be read from kernel */
	device->page_offset = PAGE_OFFSET;

	return device;
}

static int open_device_file(DumpDevice *device)
{
	int fd;

	if ((fd = open(device->device_name, O_RDONLY)) < 0)
		return FAILED;

	device->device_fd = fd;
	return SUCCEEDED;
}

static int check_fs(DumpDevice *device)
{
	struct statfs buf;
	const struct disk_dump_header *header = device->header;
	off_t memdump_size;
	bool tried_once = false;
	enum {
		check_failed = -1,
		available,
		no_space
	};

redo:
	if (fstatfs(device->file_fd, &buf) < 0)
		return check_failed;

	memdump_size = (off_t)header->written_blocks + 3; /* elf header */
	memdump_size *= header->block_size;
	memdump_size = (memdump_size + (buf.f_bsize - 1)) / buf.f_bsize;

	if (memdump_size <= buf.f_bavail)
		return available;

	if (!tried_once++)
		if (execute_script(DISKDUMP_NOSPACE, device->dump_dir) > 0)
			goto redo;

	return no_space;
}

static const char *get_dir(DumpDevice *device)
{
	char *path;
	const time_t timeofdeath = (time_t)device->header->timestamp.tv_sec;
	const char *ip_addr = convert_ip_addr(DISKDUMP_IP);
	struct tm *tm;
	char date[] = DATE_TIME_FORMAT;

	tm = localtime(&timeofdeath);
	strftime(date, sizeof(date), "%Y-%m-%d-%H:%M", tm);
	path = g_strdup_printf("%s%s-%s/", DISKDUMP_DIR_PREFIX, ip_addr, date);

	if (frc_flg) {
		if (rm_dir(path) < 0) {
			g_free(path);
			return NULL;
		}
	}

	if (mkdir(path, S_IRWXU) < 0) {
		g_free(path);
		return NULL;
	}

	return path;
}

static int open_dump_file(DumpDevice *device)
{
	const char *dir;
	char *file_name;
	int fd;

	if ((dir = get_dir(device)) == NULL)
		return FAILED;

	device->dump_dir = dir;

	file_name = g_strconcat(device->dump_dir, INCOMPLETE_DUMP_FILE, NULL);
	fd = open(file_name, O_WRONLY | O_TRUNC | O_CREAT, S_IRUSR | S_IWUSR);
	if (fd < 0) {
		g_free(file_name);
		return FAILED;
	}

	device->file_fd		= fd;
	device->file_name	= file_name;

	return SUCCEEDED;
}

static inline int get_bit(char *map, int byte, int bit)
{
	return map[byte] & (1<<bit);
}

static inline void set_bit(char *map, int byte, int bit)
{
	map[byte] |= 1<<bit;
}

static int next_bit(DumpDevice *device)
{
	char *src = device->bitmap + device->bitmap_len / 2;
	int val;

	val = get_bit(src, device->byte, device->bit);
	if (++device->bit == 8) {
		device->byte++;
		device->bit = 0;
	}

	return val;
}

static void expand_bitmap(DumpDevice *device)
{
	int byte, bit;

	for (byte = 0; byte < device->bitmap_len / 2; byte++)
		for (bit = 0; bit < 8; bit++)
			if (get_bit(device->bitmap, byte, bit)
			 && next_bit(device))
				set_bit(device->dumpable_bitmap, byte, bit);
}

static int read_device_header(DumpDevice *device)
{
	struct disk_dump_header *header;
	struct disk_dump_sub_header *sub_header;
	int ret;
	int bitmap_len;
	const int block_size = (int)sysconf(_SC_PAGESIZE);
	int sub_header_len;
	off_t offset;
	const off_t failed = (off_t)-1;

	if (block_size < 0)
		return BLOCK_SIZE_INVALID;

	header = g_malloc(block_size);

	offset = (off_t)block_size;
	if (lseek(device->device_fd, offset, SEEK_SET) == failed) {
		ret = SEEK_ERROR_IN_DUMP_HEADER;
		goto err;
	}

	if (read(device->device_fd, header, block_size) < block_size) {
		ret = READ_ERROR_IN_DUMP_HEADER;
		goto err;
	}

	/* validate dump header */
	if (memcmp(header->signature, DISK_DUMP_SIGNATURE,
				sizeof(header->signature))) {
		ret = NO_PANIC_DUMP_HEADER;
		goto err;
	}

	if (header->block_size != block_size) {
		ret = SIZE_NOT_MATCH;
		goto err;
	}

	if (sizeof(*header) + sizeof(void *) * header->nr_cpus > block_size
	 || header->nr_cpus <= 0) {
		nr_cpus = header->nr_cpus;
		ret = NUMBER_OF_CPUS_INVALID;
		goto err;
	}

	/* read sub header */
	offset = (off_t)block_size * 2;
	if (lseek(device->device_fd, offset, SEEK_SET) == failed) {
		ret = SEEK_ERROR_IN_DUMP_SUB_HEADER;
		goto err;
	}

	sub_header_len = sub_header_length(header);
	sub_header = g_malloc(sub_header_len);
	if (read(device->device_fd, sub_header, sub_header_len)
	  < sub_header_len) {
		ret = READ_ERROR_IN_DUMP_SUB_HEADER;
		goto err;
	}

	device->sub_header = sub_header;

	/* read memory bitmap */
	bitmap_len = block_size * header->bitmap_blocks;
	device->bitmap_len = bitmap_len;

	offset = (off_t)block_size * (2 + header->sub_hdr_size);
	if (lseek(device->device_fd, offset, SEEK_SET) == failed) {
		ret = SEEK_ERROR_IN_MEMORY_BITMAP;
		goto err;
	}

	device->bitmap = g_malloc(bitmap_len);
	device->dumpable_bitmap = g_malloc0(bitmap_len);
	if (read(device->device_fd, device->bitmap, bitmap_len) < bitmap_len) {
		ret = READ_ERROR_IN_MEMORY_BITMAP;
		goto err;
	}

	if (dump_is_partial(header))
		expand_bitmap(device);
	else
		memcpy(device->dumpable_bitmap, device->bitmap, bitmap_len);

	/*
	 * page 0: device header
	 * page 1: dump header
	 * page 2..: dump sub header
	 * page 2 + size of subheader: memory page 0
	 */
	device->device_page_start
		= (2 + header->sub_hdr_size + header->bitmap_blocks)
		* header->block_size;

	device->header = header;

	return COMPLETE;

err:
	g_free(header);
	return ret;
}

static bool pages_are_zero(int nr, int len, DumpDevice *device)
{
	int i;
	const unsigned int maxnr = nr + min(len, device->header->max_mapnr);

	for (i = nr >> 3; i < (maxnr + 7) >> 3; i++)
		if (device->bitmap[i])
			return false;
	return true;
}

static void build_segments_info(DumpDevice *device)
{
	const struct disk_dump_header *header = device->header;
	struct mem_segment	*seg, *seg_base;
	unsigned int chunk;			/* = HOLE_SIZE blocks */
	const unsigned int max_mapnr = header->max_mapnr;
	const unsigned int max_segments = max_mapnr / HOLE_SIZE / 2;
	bool in_segment;

	seg = seg_base = g_new0(struct mem_segment, max_segments);

	if (force_single_ptload || header->max_mapnr <= HOLE_SIZE) {
		seg->blk = 0;
		seg->len = device->header->max_mapnr;

		device->num_segments = 1;
		device->segments = seg_base;

		return;
	}

	/* The first segment must have the first block(blk#0). */
	seg = seg_base;
	seg->blk = 0;
	in_segment = true;
	for (chunk = 1; chunk < divideup(max_mapnr, HOLE_SIZE); chunk++) {
		if (pages_are_zero(chunk * HOLE_SIZE, HOLE_SIZE, device)) {
			if (in_segment) {
				seg->len = chunk * HOLE_SIZE - seg->blk;
				in_segment = false;
			}
		} else {
			if (!in_segment) {
				seg++;
				seg->blk = chunk * HOLE_SIZE;
				in_segment = true;
			}
		}
	}
	/*
	 * The last segment must contain the last blocks even if
	 * they are zero-filled pages.
	 */
	seg->len = max_mapnr - seg->blk;

	device->num_segments = seg + 1 - seg_base;
	device->segments = seg_base;
}

static bool page_is_in_segs(unsigned int nr, DumpDevice *device)
{
	struct mem_segment *seg;
	int i;

	for (i = 0; i < device->num_segments; i++) {
		seg = &device->segments[i];

		if (seg->blk <= nr && nr < seg->blk + seg->len)
			return true;
	}
	return false;
}

/******************************************************************************
 *                       Elf core dumping                                     *
 ******************************************************************************/

static inline int write_dump_header(DumpDevice *device)
{
	if (generate_elf_header(device))
		return FAILED;
	else
		return SUCCEEDED;
}

static inline bool check_zero_page(DumpDevice *device, char *data)
{
	int i;

	for (i = 0; i < device->header->block_size; i++)
		if (data[i])
			return FALSE;
	return TRUE;
}

static inline bool page_is_ram(unsigned int nr, DumpDevice *device)
{
	return get_bit(device->bitmap, nr >> 3, nr & 7);
}

static inline bool page_is_dumpable(unsigned int nr, DumpDevice *device)
{
	return device->dumpable_bitmap[nr>>3] & (1 << (nr & 7));
}

static int write_pages(DumpDevice *device)
{
	int i;
	char *data;
	int ret = COMPLETE;
	const struct disk_dump_header *header = device->header;
	const int block_size = header->block_size;
	const char *device_name = device->device_name;
	off_t offset;
	const off_t failed = (off_t)-1;

	data = g_malloc(block_size);

	offset = (off_t)device->device_page_start;
	if (lseek(device->device_fd, offset, SEEK_SET) == failed) {
		ret = SEEK_ERROR_IN_DUMP_DEVICE;
		goto out;
	}

	offset = (off_t)device->data_offset;
	if (lseek(device->file_fd, offset, SEEK_SET) == failed) {
		ret = SEEK_ERROR_IN_DUMP_FILE;
		goto out;
	}

	for (i = 0; i < header->max_mapnr; i++) {
		if (prog_flg)
			progress_bar(device_name, symbol, i, header->max_mapnr);

		if (!page_is_in_segs(i, device))
			continue;

		if (page_is_dumpable(i, device)) {
			if (read(device->device_fd, data, block_size)
			  < block_size) {
				ret = READ_ERROR_IN_DUMP_DEVICE;
				goto out;
			}
			if (check_zero_page(device, data))
				goto skip;

			if (write(device->file_fd, data, block_size)
			  < block_size) {
				ret = WRITE_ERROR_IN_DUMP_FILE;
				goto out;
			}
		} else {
skip:
			if (lseek(device->file_fd, block_size, SEEK_CUR)
			 == failed) {
				ret = SEEK_ERROR_IN_DUMP_FILE;
				goto out;
			}
		}
	}

	/*
	 * Set the crash dump file size intentionally because the last block
	 * may not be actually written.
	 */
	if ((offset = lseek(device->file_fd, 0, SEEK_CUR)) == failed) {
		ret = SEEK_ERROR_IN_DUMP_FILE;
		goto out;
	}

	if (ftruncate(device->file_fd, offset) < 0) {
		ret = DUMP_FILE_SIZE_UNDETERMINED;
		goto out;
	}

	if (prog_flg)
		progress_bar_post(device_name, symbol, i, header->max_mapnr);

out:
	g_free(data);
	return ret;
}

static int rename_dump_file(DumpDevice *device)
{
	char *old_file;
	char *new_file;
	int ret;

	old_file = g_strconcat(device->dump_dir, INCOMPLETE_DUMP_FILE, NULL);
	new_file = g_strconcat(device->dump_dir, COMPLETE_DUMP_FILE, NULL);

	ret = rename(old_file, new_file);
	device->file_name = g_strdup(new_file);

	g_free(old_file);
	g_free(new_file);

	return ret;
}

static int blank_out_signature(DumpDevice *device)
{
	int fd;
	off_t offset;
	const struct disk_dump_header *header = device->header;
	const int block_size = header->block_size;

	/* reopen device file */
	if (close(device->device_fd) < 0)
		return FAILED;
	if ((fd = open(device->device_name, O_RDWR)) < 0)
		return FAILED;
	device->device_fd = fd;

	offset = (off_t)block_size;
	if (lseek(device->device_fd, offset, SEEK_SET) < 0)
		return FAILED;

	if (write(device->device_fd, ALREADY_SAVED_SIGNATURE, SIG_LEN) < SIG_LEN)
		return FAILED;

	return SUCCEEDED;
}

int main(int argc, char *argv[])
{
	DumpDevice *device;
	int opt;
	bool verbose = false;
	int ret;

	while ((opt = getopt(argc, argv, "DfpvV")) != -1) {
		switch (opt) {
		case 'D':
			del_flg = true;
			break;
		case 'f':
			frc_flg = true;
			break;
		case 'p':
			prog_flg = true;
			break;
		case 'v':
			verbose = true;
			break;
		case 'V':
			version(CMD);
			break;
		default:
			usage(CMD);
			break;
		}
	}

	if (optind + 1 != argc)
		usage(CMD);

	if ((device = create_new_device(argv[optind])) == NULL) {
		Err("unable to create a new device.\n");
		exit(1);
	}

	if (open_device_file(device) < 0) {
		Syserr("open %s", device->device_name);
		exit(1);
	}
	Info("%s\n", device->device_name);

	ret = read_device_header(device);
	switch (ret) {
	case COMPLETE:
		break;
	case BLOCK_SIZE_INVALID:
		Err("block size in the dump header is invalid.\n");
		exit(1);
	case SEEK_ERROR_IN_DUMP_HEADER:
		Syserr("seek the dump header");
		exit(1);
	case READ_ERROR_IN_DUMP_HEADER:
		Syserr("read the dump header");
		exit(1);
	case NO_PANIC_DUMP_HEADER:
		Err("%s does not have panic dump header.\n",
				device->device_name);
		exit(1);
	case SIZE_NOT_MATCH:
		Err("block size in the dump header does not match with system page size.\n");
		exit(1);
	case NUMBER_OF_CPUS_INVALID:
		Err("invalid nr_cpus value: %d\n", nr_cpus);
		exit(1);
	case SEEK_ERROR_IN_DUMP_SUB_HEADER:
		Syserr("seek the dump sub header");
		exit(1);
	case READ_ERROR_IN_DUMP_SUB_HEADER:
		Syserr("read the dump sub header");
		exit(1);
	case SEEK_ERROR_IN_MEMORY_BITMAP:
		Syserr("seek the memory bitmap");
		exit(1);
	case READ_ERROR_IN_MEMORY_BITMAP:
		Syserr("read the memory bitmap");
		exit(1);
	default:
		Err("unknown error (reading the device header)\n");
		exit(1);
	}
	Info("  device header : read\n");

	if (open_dump_file(device) < 0) {
		Syserr("open the crash dump file");
		exit(1);
	}
	Info("  dump file name: %s\n", device->file_name);

	ret = check_fs(device);
	switch (ret) {
	case -1:
		Err("unable to check %s.\n", device->file_name);
		exit(1);
	case 0:
		break;
	case 1:
		Err("no space for dump image.\n");
		exit(1);
	default:
		Err("unknown error (checking file system)\n");
		exit(1);
	}
	Info("  file system   : available\n");

	build_segments_info(device);
	Info("  segment count : %d\n", device->num_segments);

	if (write_dump_header(device) < 0) {
		Err("unable to write the dump header.\n");
		exit(1);
	}
	Info("  dump header   : written\n");

	if (!prog_flg)
		Info("  memory dump   : ");
	ret = write_pages(device);
	switch (ret) {
	case COMPLETE:
		break;
	case SEEK_ERROR_IN_DUMP_DEVICE:
		Syserr("seek the dump device");
		exit(1);
	case READ_ERROR_IN_DUMP_DEVICE:
		Syserr("read the dump device");
		exit(1);
	case WRITE_ERROR_IN_DUMP_FILE:
		Syserr("write to the crash dump file");
		exit(1);
	case SEEK_ERROR_IN_DUMP_FILE:
		Syserr("seek the crash dump file");
		exit(1);
	case DUMP_FILE_SIZE_UNDETERMINED:
		Syserr("set the crash dump file size");
		exit(1);
	default:
		Err("unknown error (writing memory dump)\n");
		exit(1);
	}
	if (!prog_flg)
		Info("written\n");
	else
		Info("  memory dump   : written\n");
	Info("  written blocks: %u\n", device->header->written_blocks);

	if (device->header->status != DUMP_HEADER_COMPLETED) {
		Err("dump partition has incomplete memory image.\n");
		exit(1);
	}

	if (del_flg) {
		if (is_swapon()) {
			Err("swap is running.\n");
			exit(1);
		}
		if (blank_out_signature(device) == FAILED) {
			Syserr("blank out dump signature");
			exit(1);
		}
		Info("  dump signature: blanked out\n");
	}

	if (rename_dump_file(device) != 0) {
		Syserr("rename the crash dump file");
		exit(1);
	}
	Info("  dump file name: %s\n", device->file_name);

	Info("saving panic dump completed.\n");
	exit(0);
}
